/*
 * Cr le 3 mai 2005
 */
package csp_runtime;

import java.util.Vector;
import java.util.concurrent.Semaphore;

/**
 * la classe Controleur gre toutes les interactions entre processus (synchro et
 * communication).
 * 
 * @author Benjamin LE BOZEC
 */
public class Controleur {

	private Semaphore mutexList = new Semaphore (1);
		
	/**
	 * Gre les commandes d'e/s mises par un processus CSP et effectue les communications.
	 * 
	 * @param source le processus source des commandes 
	 * @param commands les commandes  traiter
	 */ 
	public void manageCommands(CspThread source, Garde... commands) {
		Vector<DebugObj> listeDebug = new Vector<DebugObj> ();
		
		/* debug */
		if (commands.length>1) {
			for (Garde cmd : commands) {
				listeDebug.add (cmd.getDebugObj ());
			}
			
			source.debug(DebugObj.toArray(listeDebug));
		}
		// on reserve l'acces
		mutexList.acquireUninterruptibly();
		
		// rcupration du ctrl associe a la source de la commande
		ThreadCtrl sourceCtrl = source.ctrl;
		
		// Recherche des commandes pouvant tre excutes (cad dont la garde
		// est bonne (=true) et dont la cmde d'i/o est executable)
		Vector<Garde> execCmds = getExecutableCommands (commands);
		
		if (execCmds.size()==0) {
			// si aucune commandes n'est executable, on vrifie qu'au moins
			// un processus est vivant avant de se mettre en attente
			Vector <Garde> toAdd = new Vector<Garde> ();
			for (Garde cmd : commands) {
				if ((cmd.bool==null) || (cmd.bool!=null && cmd.bool.eval()))
					if (cmd.ioCmd!=null && (!cmd.ioCmd.dest.ctrl.dead))
						toAdd.add (cmd);
			}
			if (toAdd.size()>0) {
				sourceCtrl.setCommandList(toAdd);
				mutexList.release();
				
				/* debug */
				if (commands.length>1) {
					listeDebug.removeAllElements();
					for (Garde cmd : toAdd)
						listeDebug.add(cmd.ioCmd.debugObj);
					source.debug(DebugObj.toArray(listeDebug));
				}
				
				try {
					sourceCtrl.sem.acquire();
				} catch (InterruptedException e) {
				}
			} else {
				mutexList.release ();
			}
		} else {
			// sinon, on choisit une commande au hasard
			Garde cmd = execCmds.get((int)(Math.random()*execCmds.size()));
			
			/* debug */
			if (commands.length>1) source.debug (cmd.getDebugObj());
			
			// on indique que la commande a t choisie
			cmd.choosen = true;
			if (cmd.ioCmd!=null) {
				// si une commande d'i/o est associ, on rcupe le 
				// proc destination, puis on effectue la communication,
				// on retire toutes ses commandes, 
				// et enfin on le libre.
				ThreadCtrl destCtrl = cmd.ioCmd.dest.ctrl;
				Garde otherCmd = destCtrl.findMatch(cmd);
				cmd.ioCmd.perform (otherCmd.ioCmd);
				otherCmd.choosen = true;
				destCtrl.commands.removeAllElements();
				destCtrl.sem.release();
			}
			
			mutexList.release();
		}
	}

	/**
	 * 
	 * @param commands
	 * @return
	 */
	private Vector<Garde> getExecutableCommands(Garde[] commands) {
		Vector<Garde> list = new Vector<Garde>();
		
		for (Garde cmd : commands) {
			if ((cmd.bool==null) || (cmd.bool!=null && cmd.bool.eval())) {
				if (cmd.ioCmd==null)
					// si la commande est une simple garde, on l'ajoute simplement
					list.add (cmd);
				else {
					// sinon, on recherche le proc de la comande d'i/o,
					// et on teste voir s'il est en attente
					ThreadCtrl destCtrl = cmd.ioCmd.dest.ctrl;
					if (destCtrl.commands.size()!=0) {
						// le proc destination est en attente
						if (destCtrl.commands.size()==1 || commands.length==1) {
							// on vrifie qu'au moins une des deux cmds n'est pas dans une alternative
							// puis on teste voir si une des commandes correspond
							if (destCtrl.findMatch(cmd)!=null) {
								list.add (cmd);
							}
						}
					}
				}
			}
		}
		
		return list;
	}
	
	/**
	 * Dclare la mort d'un processus, impliquant donc l'chec de toutes les commandes
	 * en attente de ce processus.
	 * 
	 * @param thread le processus mort.
	 */
	public void fireDeath(CspThread thread) {
		mutexList.acquireUninterruptibly();
		thread.ctrl.dead = true;
		for (CspThread th : CspThread.threadList) {
			if (th.ctrl.commands.size()>0) {
				Garde [] tab = new Garde [th.ctrl.commands.size()];
				th.ctrl.commands.toArray(tab);
				for (Garde cmd : tab) {
					if (cmd.ioCmd!=null && cmd.ioCmd.dest==thread) {
						th.ctrl.commands.remove(cmd);
					}
				}
				if  (th.ctrl.commands.size()==0) th.ctrl.sem.release();
			}
		}
		mutexList.release ();
	}

}