/*
 * Cr le 02/05/05
 * 
 * Modifications:
 * - 09/05/05 : changement des procdures "input" et "output" ('privatisation') 
 * 				AJout de la mthode getNewSelect
 */
package csp_runtime;

import static csp_runtime.IOType.*;

import java.awt.Point;
import java.util.*;

/**
 * CspThread correspond  la super classe de tous les threads CSP. Elle fournit 
 * des mthodes standards telles que "input" et "output" pour les communication, 
 * ainsi que "fireDeath" pour dclarer la mort d'un processus.
 *  
 * @author Benjamin LE BOZEC
 */
public abstract class CspThread implements Runnable {
	
	protected static final Collection<CspThread> threadList = Collections.synchronizedCollection(new Vector<CspThread>()); 
	
	/**
	 * controleur charg de grer les communications
	 */
	public static final Controleur controleur = new Controleur ();
	
	/**
	 * Objet utilis par le controleur pour grer les communicationss
	 */
	protected final ThreadCtrl ctrl = new ThreadCtrl ();
	
	private static Collection<NewThreadListener> newThreadListeners = Collections.synchronizedCollection(new Vector<NewThreadListener>());	
	private static Collection<EndThreadListener> endThreadListeners = Collections.synchronizedCollection(new Vector<EndThreadListener>());
	
	/**
	 * Si sortieConsoleStd=true, la mthode print utilise System.out.print ()
	 */
	private static boolean sortieConsoleStd = true;
	
	/**
	 * Le thread associ au processus CSP
	 */
	private Thread thread;
	
	/**
	 * Nom du processus
	 */
	private String name = "";

	/**
	 * boolean indiquant si le processus a chou ou non
	 */
	protected boolean success = false;
	
	private Collection<PrintListener> printListeners = new Vector<PrintListener>();
	private Collection<DebugListener> debugListeners = new Vector<DebugListener>();
	private Collection<DebugListener> blockListeners = new Vector<DebugListener>();
	
	private Point firstPoint = new Point (0,0);
	private Point lastPoint = new Point (0,0);

	private String consoleBuffer = "";
	private DebugObj [] lastDebugObjs = null;
	
	private boolean mustStop = false;

	private String completeName;
	
	/**
	 * Constructeur par dfaut.
	 */
	public CspThread () {
		threadList.add(this);
	}
	
	public void finalize () {
		threadList.remove (this);
	}
	
	/**
	 * Constructeur.
	 * @param name le nom du processus
	 */
	public CspThread (String name) {
		this();
		this.name = name;
	}

	/**
	 * input correspond  la commande d'entre de CSP.
	 * 
	 * @param cible le processus cible de la commande (celui dont les donnes doivent tre rcupres)
	 * @param expr l'expression cible
	 * @return vrai si la communication a t un succs, faux si la commande a chou
	 */
	public boolean input (CspThread cible, Expression expr) {
		Garde cmd = new Garde (null, new IOCommand (this, cible, INPUT, expr));
		controleur.manageCommands (this, cmd); 
		return cmd.getSuccess();
	}
	
	/**
	 * output correspond  la commande de sortie de CSP
	 * 
	 * @param cible le processus cible de la commande (a qui 'expr' doit tre envoye)
	 * @param expr l'expression a transmettre
	 * @return VRAI si succs, FAUX si la commande a chou
	 */
	public boolean output (CspThread cible, Expression expr) {
		Garde cmd = new Garde (null, new IOCommand (this, cible, OUTPUT, expr));
		controleur.manageCommands (this, cmd);
		return cmd.getSuccess();
	}

	/**
	 * getNewSelect initialise et retourne un nouvel objet Select associ avec le
	 * processus courant.
	 * 
	 * @return un objet Select
	 */
	public Select getNewSelect () {
		return new Select (this);
	}
	
	/**
	 * fireDeath() est appel par le processus courant pour signaler au controleur
	 * sa mort. 
	 */
	public void fireDeath () {
		debug ((DebugObj [])null);
		clear ();
		fireDeath (this);
	}
	
	private void fireDeath(CspThread thread) {
		controleur.fireDeath(thread);
		
		for (EndThreadListener listener : endThreadListeners)
			listener.endThread (thread);
	}

	/**
	 * start() lance le processus
	 */
	public void start () {
		for (NewThreadListener listener : newThreadListeners)
			listener.newThread(this);
	
		thread = new Thread (this);
		thread.start();
	}
	
	/**
	 * @return le nom du processus
	 */
	public String getName() {
		if (name.equals("")) {
			return getClass().getSimpleName();
		} else 
			return name;
	}

	/**
	 * join attend la fin du processus.
	 * @return FAUX si le processus a chou, VRAI sinon
	 */
	public boolean join () {
		try {
			thread.join();
		} catch (InterruptedException e) {
		}
		return success;
	}
	
	public void print (Object obj) {
		consoleBuffer += obj.toString ();
		for (PrintListener listener : printListeners)
			listener.print(obj.toString());
		if (sortieConsoleStd)
			System.out.print (obj);
	}
	
	public String getConsoleBuffer () {
		return consoleBuffer;
	}

	public void addPrintListener (PrintListener listener) {
		printListeners.add (listener);
	}
	
	public void removePrintListener (PrintListener listener) {
		printListeners.remove (listener);
	}
	
	public static void addNewThreadListener (NewThreadListener listener) {
		newThreadListeners.add (listener);
	}
	
	public static void removeNewThreadListener (NewThreadListener listener){
		newThreadListeners.remove (listener);
	}

	public static void addEndThreadListener (EndThreadListener listener) {
		endThreadListeners.add (listener);
	}
	
	public static void removeEndThreadListener (EndThreadListener listener){
		endThreadListeners.remove (listener);
	}

	public String toString () {
		return getName();
	}

	protected void setThreadBounds (int l1, int c1, int l2, int c2) {
		this.firstPoint = new Point (c1, l1);
		this.lastPoint = new Point (c2, l2);
	}
	
	public Point getFirstPoint () {
		return firstPoint;
	}
	
	public Point getLastPoint () {
		return lastPoint;
	}
	
	public void addDebugListener (DebugListener listener) {
		synchronized (debugListeners) {
			debugListeners.add (listener);
		}
	}
	
	public void removeDebugListener (DebugListener listener) {
		synchronized (debugListeners) {
			debugListeners.remove (listener);
		}
	}

	public void addBlockListener (DebugListener listener) {
		synchronized (blockListeners) {
			blockListeners.add (listener);
		}
	}
	
	public void removeBlockListener (DebugListener listener) {
		synchronized (blockListeners) {
			blockListeners.remove (listener);
		}
	}

	/**
	 * @return vrai si le proc doit s'arreter
	 */
	public boolean debug (DebugObj ... debugObjs) {
		lastDebugObjs = debugObjs;
		synchronized (debugListeners) {
			for (DebugListener listener : debugListeners)
				listener.debug (debugObjs);
		}
		if (getMustStop()) return true;
		synchronized (blockListeners) {
			for (DebugListener listener : blockListeners)
				listener.debug (debugObjs);
		}
		return false;
	}
	
	public synchronized boolean getMustStop () {
		return mustStop;
	}
	
	public synchronized void setMustTop (boolean b) {
		mustStop  = b;
	}
	
	public static void terminateAll () {
		for (CspThread thread : threadList) {
			thread.terminate();
		}
	}

	private void terminate() {
		if (thread != null) {
			setMustTop(true);
			thread.interrupt();
			//fireDeath ();
		}
	}

	protected void setCompleteName (int ... indices) {
		completeName = getName ();
		if (indices.length>0) {
			completeName += "[";
			for (int i=0;i<indices.length;i++) {
				if (i!=0) completeName += ", ";
				completeName += indices[i];
			}
			completeName += "]";
		}
	}
	
	public String getCompleteName() {
		return completeName;
	}
	
	public DebugObj [] getLastDebugObjs () {
		return lastDebugObjs;
	}
	
	public void clear () {
		printListeners.clear();
		debugListeners.clear();
		blockListeners.clear();
	}

	public static void flushEverything() {
		for (CspThread thread : threadList)
			thread.clear();
		threadList.clear();
		newThreadListeners.clear();
		endThreadListeners.clear();
	}
	
	/**
	 * Si b = vrai, les sorties des processus (appel  la mthode print) 
	 * seront rpercutes sur la console standard (System.out.print)
	 */
	public static void setSortieConsoleStd (boolean b) {
		sortieConsoleStd = b;
	}
}


